(*
    Stack Quizzer - A tool for memorizing decks of cards
    Interface to CARDS.DLL (16/32 bit)
    Copyright (c) 2003 Alessandro Scotti
    http://www.ascotti.org/

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*)
unit Cards;

interface

uses
  Graphics;

const
  BackCrossHatch = 52;
  BackBlack = 53;
  BackWeave1 = 54;
  BackWeave2 = 55;
  BackRobot = 56;
  BackFlowers = 57;
  BackVine1 = 58;
  BackVine2 = 59;
  BackFish1 = 60;
  BackFish2 = 61;
  BackShells = 62;
  BackCastle = 63;
  BackIsland = 64;
  BackCardHand = 65;
  BackUnused = 66;
  BackTheX = 67;
  BackTheO = 68;

const
  SuitClubs = 0;
  SuitDiamonds = 1;
  SuitHearts = 2;
  SuitSpades = 3;

function cardMake( AValue, ASuit: integer ): integer;
procedure cardUnmake( ACard: integer; var AValue, ASuit: integer );
procedure cardDraw( ACanvas: TCanvas; AValue, AX, AY: integer );
procedure cardDrawEx( ACanvas: TCanvas; AValue, AX, AY, AKind: integer; AColor: TColor );
function cardGetDefaultWidth: integer;
function cardGetDefaultHeight: integer;

implementation

uses Windows;

const
  CardFace = 0;
  CardBack = 1;
  CardInverted = 2;

const
  MaxValue = BackTheO;

const
  DllInitCount: integer = 0;
  DllName = 'CARDS.DLL';

type
  TCdtInit = function( var Width, Height: integer ): LongBool; stdcall;
  TCdtDraw = function( DC: HDC; X, Y: integer; Card, Kind: integer; Color: cardinal ): LongBool; stdcall;
  TCdtDrawEx = function( DC: HDC; X, Y, DX, DY: integer; Card, Kind: integer; Color: cardinal ): LongBool; stdcall;
  TCdtAnimate = function( DC: HDC; Back: integer; X, Y: integer; Frame: integer ): LongBool; stdcall;
  TCdtTerm = procedure; stdcall;

var
  DllHandle: THandle;
  DllCardWidth: integer;
  DllCardHeight: integer;
  CdtInit: TCdtInit;
  CdtDraw: TCdtDraw;
  CdtDrawEx: TCdtDrawEx;
  CdtAnimate: TCdtAnimate;
  CdtTerm: TCdtTerm;

function IsCardLibraryInitialized: boolean;
begin
  Result := DllInitCount > 0;
end;

// Functions from WOW32 not declared in Delphi
function WOWGlobalAllocLock16( Flags: LongWord; Size: dword; var Handle: word ): pointer; stdcall; external 'WOW32.DLL';
function WOWGetVDMPointerFix( PMem16: pointer; Size: dword; ProtMode: LongBool ): pointer; stdcall; external 'WOW32.DLL';
function WOWGlobalUnlockFree16( PMem16: pointer ): LongWord; stdcall; external 'WOW32.DLL';
function WOWHandle16( Handle: THandle; Kind: integer ): LongWord; stdcall; external 'WOW32.DLL';

// Undocumented kernel functions, imported by ordinal
function LoadLibrary16( Name: PChar ): THandle; stdcall; external 'KERNEL32.DLL' index 35;
procedure FreeLibrary16( Handle: THandle ); stdcall; external 'KERNEL32.DLL' index 36;
function GetProcAddress16( Handle: THandle; Name: PChar ): FARPROC; stdcall; external 'KERNEL32.DLL' index 37;

const
  WOW_TYPE_HWND = 0;
  WOW_TYPE_HMENU =      WOW_TYPE_HWND + 1;
  WOW_TYPE_HDWP =       WOW_TYPE_HWND + 2;
  WOW_TYPE_HDROP =      WOW_TYPE_HWND + 3;
  WOW_TYPE_HDC =        WOW_TYPE_HWND + 4;
  WOW_TYPE_HFONT =      WOW_TYPE_HWND + 5;
  WOW_TYPE_HMETAFILE =  WOW_TYPE_HWND + 6;
  WOW_TYPE_HRGN =       WOW_TYPE_HWND + 7;
  WOW_TYPE_HBITMAP =    WOW_TYPE_HWND + 8;
  WOW_TYPE_HBRUSH =     WOW_TYPE_HWND + 9;
  WOW_TYPE_HPALETTE =   WOW_TYPE_HWND + 10;
  WOW_TYPE_HPEN =       WOW_TYPE_HWND + 11;
  WOW_TYPE_HACCEL =     WOW_TYPE_HWND + 12;
  WOW_TYPE_HTASK =      WOW_TYPE_HWND + 13;
  WOW_TYPE_FULLHWND =   WOW_TYPE_HWND + 14;

var
  CdtInit16: FARPROC;
  CdtDraw16: FARPROC;
  CdtTerm16: FARPROC;
  QT_Thunk: FARPROC;

{$O-}

var
  HDC16: word;
  HMem16: word;
  PMem16: pointer;
  PMem32: ^integer;

function CdtInitProxy( var Width, Height: integer ): LongBool; stdcall;
var
  Stack: array[1..64] of byte;
begin
  Stack[1] := 0;

  PMem16 := WOWGlobalAllocLock16( GMEM_FIXED or GMEM_ZEROINIT or GMEM_SHARE, 32, HMem16 );
  PMem32 := WOWGetVDMPointerFix( PMem16, 32, true );

  asm
    mov eax, [PMem16]
    push eax
    add eax, 4
    push eax
    mov edx, [CdtInit16]
    call QT_Thunk
    and eax, $FFFF
    mov [Result], eax
  end;

  DllCardWidth := PMem32^; Inc(PMem32);
  DllCardHeight := PMem32^;

  WOWGlobalUnlockFree16( PMem16 );
end;

function CdtDrawProxy( DC: HDC; X, Y: integer; Card, Kind: integer; Color: cardinal ): LongBool; stdcall;
var
  Stack: array[1..64] of byte;
begin
  Stack[1] := 0;

  HDC16 := WOWHandle16( DC, WOW_TYPE_HDC );

  asm
    mov ax, [HDC16]
    push ax
    mov eax, [X]
    push ax
    mov eax, [Y]
    push ax
    mov eax, [Card]
    push ax
    mov eax, [Kind]
    push ax
    mov eax, [Color]
    push eax
    mov edx, [CdtDraw16]
    call QT_Thunk
    and eax, $FFFF
    mov [Result], eax
  end;
end;

procedure CdtTermProxy; stdcall;
var
  Stack: array[1..64] of byte;
begin
  Stack[1] := 0;

  asm
    mov edx, [CdtTerm16]
    call QT_Thunk
  end;
end;

{$O+}

function InitializeCardLibrary: boolean;
var
  HKernel: THandle;
begin
  Result := false;

  if( DllInitCount = 0 ) then begin
    DllCardWidth := 0;
    DllCardHeight := 0;

    // Load and initialize the card library
    if( (GetVersion and $80000000) <> 0 ) then begin
      // DLL is 16-bit, we have a lot of work to do...
      DllHandle := LoadLibrary16( DllName );
      if( DllHandle <= 32 )then
        Exit;

      // Get the address of the thunking function, use name
      // lookup as ordinal can be different across systems
      HKernel := GetModuleHandle( 'KERNEL32' );
      QT_Thunk := GetProcAddress( HKernel, 'QT_Thunk' );

      // Get the addresses of the 16-bit functions we need
      CdtInit16 := GetProcAddress16(DllHandle, 'cdtInit');
      CdtDraw16 := GetProcAddress16(DllHandle, 'cdtDraw');
      CdtTerm16 := GetProcAddress16(DllHandle, 'cdtTerm');

      // Redirect the standard (32-bit) functions to our
      // proxies that will thunk to the 16-bit code
      CdtInit := CdtInitProxy;
      CdtDraw := CdtDrawProxy;
      cdtTerm := CdtTermProxy;
    end
    else begin
      // DLL is 32-bit... this is much easier!
      DllHandle := LoadLibrary( DllName );
      if( DllHandle = 0 ) then
        Exit;

      @CdtInit := GetProcAddress( DllHandle, 'cdtInit' );
      @CdtDraw := GetProcAddress( DllHandle, 'cdtDraw' );
      @CdtDrawEx := GetProcAddress( DllHandle, 'cdtDrawEx' );
      @CdtAnimate := GetProcAddress( DllHandle, 'cdtAnimate' );
      @CdtTerm := GetProcAddress( DllHandle, 'cdtTerm' );
    end;

    CdtInit( DllCardWidth, DllCardHeight );
  end;

  Inc( DllInitCount );

  Result := true;
end;

procedure TerminateCardLibrary;
begin
  Dec( DllInitCount );

  if( DllInitCount = 0 ) then begin
    // Unload library
    CdtTerm;
    FreeLibrary( DllHandle );
  end;
end;

function cardMake( AValue, ASuit: integer ): integer;
begin
  Result := (AValue-1)*4 + ASuit;
end;

procedure cardUnmake( ACard: integer; var AValue, ASuit: integer );
begin
  ASuit := ACard mod 4;
  AValue := ACard div 4 + 1;
end;

procedure cardDraw( ACanvas: TCanvas; AValue, AX, AY: integer );
var
  Kind: integer;
begin
  if( IsCardLibraryInitialized ) then begin
    if( AValue <= MaxValue ) then begin
      Kind := CardFace;
      if( AValue > 52 ) then
        Kind := CardBack;
      CdtDraw( ACanvas.Handle, AX, AY, AValue, Kind, 0 );
    end;
  end;
end;

procedure cardDrawEx( ACanvas: TCanvas; AValue, AX, AY, AKind: integer; AColor: TColor );
begin
  if( IsCardLibraryInitialized ) then begin
    CdtDraw( ACanvas.Handle, AX, AY, AValue, AKind, AColor );
  end;
end;

function cardGetDefaultWidth: integer;
begin
  Result := DllCardWidth;
end;

function cardGetDefaultHeight: integer;
begin
  Result := DllCardHeight;
end;

initialization
  InitializeCardLibrary;

finalization
  TerminateCardLibrary;

end.
