(*
    Stack Quizzer - A tool for memorizing decks of cards
    Main window
    Copyright (c) 2003 Alessandro Scotti
    http://www.ascotti.org/

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*)
unit Main;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, Cards, ExtCtrls, Menus, ShellAPI,
  Deck;

type
  TFormMain = class(TForm)
    ComboStack: TComboBox;
    LabelStack: TLabel;
    ComboDisplayType: TComboBox;
    LabelDisplayType: TLabel;
    LabelRange: TLabel;
    EditRangeLo: TEdit;
    EditRangeHi: TEdit;
    LabelRangeTo: TLabel;
    PanelCard: TPanel;
    ButtonNext: TButton;
    ButtonRandom: TButton;
    CheckShowAll: TCheckBox;
    ButtonPrevious: TButton;
    PaintBoxCard: TPaintBox;
    MainMenu: TMainMenu;
    CmdFile: TMenuItem;
    CmdExit: TMenuItem;
    CmdHelp: TMenuItem;
    CmdAbout: TMenuItem;
    CmdCard: TMenuItem;
    CmdFirst: TMenuItem;
    CmdLast: TMenuItem;
    CmdNext: TMenuItem;
    CmdPrevious: TMenuItem;
    CmdRandom: TMenuItem;
    MenuSeparator1: TMenuItem;
    CmdResetRange: TMenuItem;
    EditAutoAnswer: TEdit;
    LabelSeconds: TLabel;
    LabelAutoAnswer: TLabel;
    TimerAutoAnswer: TTimer;
    Deck1: TMenuItem;
    CmdEditStack: TMenuItem;
    CheckNoAutoAnswer: TCheckBox;
    CmdNewStack: TMenuItem;
    CmdDeleteStack: TMenuItem;
    CheckShuffled: TCheckBox;
    MenuSeparator2: TMenuItem;
    CmdProgramHelp: TMenuItem;
    MenuSeparator3: TMenuItem;
    CmdShuffle: TMenuItem;
    procedure FormCreate(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure ButtonNextClick(Sender: TObject);
    procedure ButtonPreviousClick(Sender: TObject);
    procedure ComboStackChange(Sender: TObject);
    procedure ComboDisplayTypeChange(Sender: TObject);
    procedure ButtonRandomClick(Sender: TObject);
    procedure PaintBoxCardPaint(Sender: TObject);
    procedure CheckShowAllClick(Sender: TObject);
    procedure EditRangeHiExit(Sender: TObject);
    procedure EditRangeLoExit(Sender: TObject);
    procedure CmdFirstClick(Sender: TObject);
    procedure CmdLastClick(Sender: TObject);
    procedure CmdResetRangeClick(Sender: TObject);
    procedure CmdExitClick(Sender: TObject);
    procedure ButtonAddStackClick(Sender: TObject);
    procedure TimerAutoAnswerTimer(Sender: TObject);
    procedure EditAutoAnswerChange(Sender: TObject);
    procedure CheckNoAutoAnswerClick(Sender: TObject);
    procedure CmdNewStackClick(Sender: TObject);
    procedure CmdEditStackClick(Sender: TObject);
    procedure CmdDeleteStackClick(Sender: TObject);
    procedure CmdAboutClick(Sender: TObject);
    procedure CmdProgramHelpClick(Sender: TObject);
    procedure CheckShuffledClick(Sender: TObject);
    procedure CmdShuffleClick(Sender: TObject);
    procedure FormKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
  private
    { Private declarations }
    FDeck: TDeck;
    FCurrentIndex: integer;
    FShowFace: boolean;
    FRangeLo: integer;
    FRangeHi: integer;
    FAutoAnswerDelay: integer;
    FAutoAnswerTimer: integer;
    FAutoAnswerEvent: boolean;
    FPresetDecks: TStringList;
    FShuffleMap: array[ 1..CardsInDeck ] of integer;

    function GetCurrentCard: integer;
    function GetCurrentIndex: integer;
    procedure CheckComboStackNotEmpty;
    procedure LoadConfiguration;
    procedure PaintCurrentCard;
    procedure RangeChanged;
    procedure ResetAutoAnswer;
    procedure SaveConfiguration;
    procedure SetCurrentIndex( AIndex: integer );
    procedure Shuffle;
    procedure UpdateNextPreviousButtons;
  public
    { Public declarations }
  end;

var
  FormMain: TFormMain;

implementation

uses
  IniFiles, Editor, GetStackName, About, Help;

const
  NewDeckOrder = 'AH 2H 3H 4H 5H 6H 7H 8H 9H 10H JH QH KH AC 2C 3C 4C 5C 6C 7C 8C 9C 10C JC QC KC KD QD JD 10D 9D 8D 7D 6D 5D 4D 3D 2D AD KS QS JS 10S 9S 8S 7S 6S 5S 4S 3S 2S AS';

{$R *.DFM}

const
  IniSectionStacks = 'Stacks';
  IniSectionOptions = 'Options';
  IniValueRangeLo = 'RangeLo';
  IniValueRangeHi = 'RangeHi';
  IniValueDisplayType = 'DisplayType';
  IniValueAnswerDelay = 'AnswerDelay';
  IniValueNoDups = 'Shuffled';
  IniValueNoAnswer = 'NoAnswer';
  IniValueDisplayAll = 'DisplayAll';
  IniValueDefaultStack = 'DefaultStack';

procedure TFormMain.LoadConfiguration;
var
  F: TIniFile;
  L: TStringList;
  I, K: integer;
  N, V: string;
begin
  try
    F := TIniFile.Create( ChangeFileExt(Application.ExeName,'.ini') );

    // Load the preset decks from the initialization file
    L := TStringList.Create;
    F.ReadSectionValues( IniSectionStacks, L );
    for I:=1 to L.Count do begin
      N := L.Names[ I-1 ];
      V := L.Values[ N ];
      if( deckFromString( V, nil ) = 0 ) then begin
        K := ComboStack.Items.Add( N );
        FPresetDecks.Insert( K, V );
      end;
    end;
    L.Free;

    // Load the options
    FRangeLo := F.ReadInteger( IniSectionOptions, IniValueRangeLo, FRangeLo );
    FRangeHi := F.ReadInteger( IniSectionOptions, IniValueRangeHi, FRangeHi );
    ComboDisplayType.ItemIndex := F.ReadInteger( IniSectionOptions, IniValueDisplayType, 0 );
    ComboStack.ItemIndex := F.ReadInteger( IniSectionOptions, IniValueDefaultStack, 0 );
    FAutoAnswerDelay := F.ReadInteger( IniSectionOptions, IniValueAnswerDelay, FAutoAnswerDelay );
    CheckShuffled.Checked := F.ReadInteger( IniSectionOptions, IniValueNoDups, 0 ) <> 0;
    CheckNoAutoAnswer.Checked := F.ReadInteger( IniSectionOptions, IniValueNoAnswer, 0 ) <> 0;
    CheckShowAll.Checked := F.ReadInteger( IniSectionOptions, IniValueDisplayAll, 0 ) <> 0;

    F.Free;
  except
  end;
end;

procedure TFormMain.SaveConfiguration;
var
  F: TIniFile;
  I: integer;
function B2I( B: boolean ): integer;
  begin
    Result := 0;
    if( B ) then Result := 1;
  end;
begin
  try
    F := TIniFile.Create( ChangeFileExt(Application.ExeName,'.ini') );

    // Save the preset decks into the initialization file
    try
      F.EraseSection( IniSectionStacks );
    except
    end;

    for I:=1 to ComboStack.Items.Count do begin
      F.WriteString( IniSectionStacks, ComboStack.Items[I-1], FPresetDecks[I-1] );
    end;

    // Save the options
    F.WriteInteger( IniSectionOptions, IniValueRangeLo, FRangeLo );
    F.WriteInteger( IniSectionOptions, IniValueRangeHi, FRangeHi );
    F.WriteInteger( IniSectionOptions, IniValueDisplayType, ComboDisplayType.ItemIndex );
    F.WriteInteger( IniSectionOptions, IniValueDefaultStack, ComboStack.ItemIndex );
    F.WriteInteger( IniSectionOptions, IniValueAnswerDelay, FAutoAnswerDelay );
    F.WriteInteger( IniSectionOptions, IniValueNoDups, B2I(CheckShuffled.Checked) );
    F.WriteInteger( IniSectionOptions, IniValueNoAnswer, B2I(CheckNoAutoAnswer.Checked) );
    F.WriteInteger( IniSectionOptions, IniValueDisplayAll, B2I(CheckShowAll.Checked) );

    F.Free;
  except
  end;
end;

procedure TFormMain.FormCreate(Sender: TObject);
begin
  Randomize;

  FRangeLo := 1;
  FRangeHi := 52;
  FPresetDecks := TStringList.Create;
  FShowFace := Random(2) = 0;
  FAutoAnswerEvent := false;
  FAutoAnswerTimer := 0;
  FAutoAnswerDelay := 2000; // Milliseconds

  ComboDisplayType.ItemIndex := 0;

  LoadConfiguration;

  // Make sure range and current index are valid
  if( FRangeLo < 1 )or( FRangeLo > CardsInDeck )or( FRangeLo > FRangeHi ) then begin
    FRangeLo := 1;
  end;
  if( FRangeHi < 1 )or( FRangeHi > CardsInDeck ) then begin
    FRangeHi := CardsInDeck;
  end;
  FCurrentIndex := FRangeLo;

  // Update controls to reflect status of loaded parameters
  EditRangeLo.Text := IntToStr( FRangeLo );
  EditRangeHi.Text := IntToStr( FRangeHi );
  EditAutoAnswer.Text := IntToStr( FAutoAnswerDelay div 1000 );

  CheckComboStackNotEmpty;

  deckFromString( FPresetDecks[ComboStack.ItemIndex], @FDeck );

  Shuffle;
end;

procedure TFormMain.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  SaveConfiguration;
end;

procedure TFormMain.CheckComboStackNotEmpty;
var
  N: integer;
begin
  if( ComboStack.Items.Count = 0 ) then begin
    N := ComboStack.Items.Add( 'New Deck Order (Bicycle)' );
    FPresetDecks.Insert( N, NewDeckOrder );
    ComboStack.ItemIndex := 0;
  end;
end;

procedure TFormMain.SetCurrentIndex( AIndex: integer );
begin
  if( AIndex < FRangeLo ) then
    AIndex := FRangeLo;
  if( AIndex > FRangeHi ) then
    AIndex := FRangeHi;
  if( AIndex <> FCurrentIndex ) then begin
    FCurrentIndex := AIndex;

    FShowFace := ComboDisplayType.ItemIndex = 1;
    if( ComboDisplayType.ItemIndex = 0 ) then begin
      FShowFace := Random(2) = 0;
    end;

    UpdateNextPreviousButtons;

    PaintCurrentCard;
  end;
end;

procedure TFormMain.UpdateNextPreviousButtons;
begin
  ButtonNext.Enabled := FCurrentIndex < FRangeHi;
  CmdNext.Enabled := FCurrentIndex < FRangeHi;
  ButtonPrevious.Enabled := FCurrentIndex > FRangeLo;
  CmdPrevious.Enabled := FCurrentIndex > FRangeLo;
end;

procedure TFormMain.RangeChanged;
begin
  Shuffle;
  UpdateNextPreviousButtons;
end;

procedure TFormMain.Shuffle;
const
  SwapIterations = 50;
var
  I, J, K, L, M, R: integer;
begin
  R := FRangeHi - FRangeLo + 1;
  M := FRangeLo;

  for I:=1 to CardsInDeck do
    FShuffleMap[I] := I;

  for I:=1 to SwapIterations do begin
    J := M + Random( R );
    K := M + Random( R );
    L := FShuffleMap[J];
    FShuffleMap[J] := FShuffleMap[K];
    FShuffleMap[K] := L;
  end;
end;

function TFormMain.GetCurrentIndex: integer;
begin
  Result := FCurrentIndex;
  if( CheckShuffled.Checked ) then begin
    Result := FShuffleMap[ FCurrentIndex ];
  end;
end;

function TFormMain.GetCurrentCard: integer;
begin
  Result := FDeck[ GetCurrentIndex ];
end;

procedure TFormMain.PaintCurrentCard;
var
  Text: string;
begin
  PaintBoxCard.Canvas.FillRect( Rect(0,0,Width,Height) );

  if( CheckShowAll.Checked )or( FAutoAnswerEvent and not CheckNoAutoAnswer.Checked) then begin
    // Faces and numbers
    Text := IntToStr( GetCurrentIndex );
    PaintBoxCard.Canvas.Font.Color := clBlue;
    PaintBoxCard.Canvas.TextOut(
      (PaintBoxCard.Width div 2 - PaintBoxCard.Canvas.TextWidth(Text)) div 2,
      (PaintBoxCard.Height - PaintBoxCard.Canvas.TextHeight(Text)) div 2,
      Text );

    PaintBoxCard.Canvas.Font.Color := clBlack;
    cardDraw( PaintBoxCard.Canvas,
      GetCurrentCard,
      PaintBoxCard.Width div 2 + (PaintBoxCard.Width div 2 - cardGetDefaultWidth) div 2,
      (PaintBoxCard.Height - cardGetDefaultHeight) div 2 );
  end
  else if( FShowFace ) then begin
    // Faces
    PaintBoxCard.Canvas.Font.Color := clBlack;
    cardDraw( PaintBoxCard.Canvas,
      GetCurrentCard,
      (PaintBoxCard.Width - cardGetDefaultWidth) div 2,
      (PaintBoxCard.Height - cardGetDefaultHeight) div 2 );
  end
  else begin
    // Numbers
    Text := IntToStr( GetCurrentIndex );
    PaintBoxCard.Canvas.Font.Color := clBlue;
    PaintBoxCard.Canvas.TextOut(
      (PaintBoxCard.Width - PaintBoxCard.Canvas.TextWidth(Text)) div 2,
      (PaintBoxCard.Height - PaintBoxCard.Canvas.TextHeight(Text)) div 2,
      Text );
  end;
end;

function FindCard( var Deck: TDeck; Card: integer ): integer;
begin
  Result := 1;
  while( Result <= 52 ) do begin
    if( Card = Deck[Result] ) then
      break;
    Inc( Result );
  end;
  if( Result > 52 ) then
    Result := 0;
end;

procedure TFormMain.ButtonNextClick(Sender: TObject);
begin
  ResetAutoAnswer;
  SetCurrentIndex( FCurrentIndex+1 );
end;

procedure TFormMain.ButtonPreviousClick(Sender: TObject);
begin
  ResetAutoAnswer;
  SetCurrentIndex( FCurrentIndex-1 );
end;

procedure TFormMain.ComboStackChange(Sender: TObject);
begin
  DeckFromString( FPresetDecks[ ComboStack.ItemIndex ], @FDeck );
  FCurrentIndex := 1;
  PaintCurrentCard;
end;

procedure TFormMain.ComboDisplayTypeChange(Sender: TObject);
begin
  PaintCurrentCard;
end;

procedure TFormMain.ButtonRandomClick(Sender: TObject);
var
  I: integer;
begin
  repeat
    I := Random( FRangeHi - FRangeLo + 1 ) + FRangeLo;
  until( I <> FCurrentIndex );
  ResetAutoAnswer;
  SetCurrentIndex( I );
end;

procedure TFormMain.PaintBoxCardPaint(Sender: TObject);
begin
  PaintCurrentCard;
end;

procedure TFormMain.CheckShowAllClick(Sender: TObject);
begin
  PaintCurrentCard;
end;

procedure TFormMain.EditRangeHiExit(Sender: TObject);
var
  N: integer;
begin
  try
    N := StrToInt( EditRangeHi.Text );
    if( N >= 1 )and( N <= 52 )and( N >= FRangeLo ) then begin
      FRangeHi := N;
      RangeChanged;
    end;
  except
  end;
  EditRangeHi.Text := IntToStr(FRangeHi);
end;

procedure TFormMain.EditRangeLoExit(Sender: TObject);
var
  N: integer;
begin
  try
    N := StrToInt( EditRangeLo.Text );
    if( N >= 1 )and( N <= 52 )and( N <= FRangeHi ) then begin
      FRangeLo := N;
      RangeChanged;
    end;
  except
  end;
  EditRangeLo.Text := IntToStr(FRangeLo);
end;

procedure TFormMain.CmdFirstClick(Sender: TObject);
begin
  ResetAutoAnswer;
  Shuffle;
  SetCurrentIndex( FRangeLo );
end;

procedure TFormMain.CmdLastClick(Sender: TObject);
begin
  ResetAutoAnswer;
  SetCurrentIndex( FRangeHi );
end;

procedure TFormMain.CmdResetRangeClick(Sender: TObject);
begin
  FRangeLo := 1;
  FRangeHi := 52;
  EditRangeLo.Text := '1';
  EditRangeHi.Text := '52';
  RangeChanged;
end;

procedure TFormMain.CmdExitClick(Sender: TObject);
begin
  Close;
end;

procedure TFormMain.ButtonAddStackClick(Sender: TObject);
begin
  FormEditor.ShowModal;
end;

procedure TFormMain.ResetAutoAnswer;
begin
  FAutoAnswerEvent := false;
  FAutoAnswerTimer := 0;
  TimerAutoAnswer.Enabled := true;
end;

procedure TFormMain.TimerAutoAnswerTimer(Sender: TObject);
begin
  Inc( FAutoAnswerTimer, TimerAutoAnswer.Interval );
  if( FAutoAnswerTimer > FAutoAnswerDelay ) then begin
    TimerAutoAnswer.Enabled := false;
    FAutoAnswerEvent := true;
    PaintCurrentCard;
  end;
end;

procedure TFormMain.EditAutoAnswerChange(Sender: TObject);
var
  N: integer;
begin
  try
    N := StrToInt( EditAutoAnswer.Text );
    if( N >= 1 )and( N <= 10 ) then begin
      FAutoAnswerDelay := N*1000;
    end;
  except
  end;
end;

procedure TFormMain.CheckNoAutoAnswerClick(Sender: TObject);
begin
  PaintCurrentCard;
end;

procedure TFormMain.CmdNewStackClick(Sender: TObject);
var
  Name: string;
  Done: boolean;
  Deck: TDeck;
  N: integer;
begin
  FormGetStackName := TFormGetStackName.Create( Self );
  repeat
    Done := true;
    if( FormGetStackName.ShowModal = mrOk ) then begin
      Name := FormGetStackName.EditName.Text;
      if( ComboStack.Items.IndexOf( Name ) >= 0 ) then begin
        Done := MessageDlg( 'There is already a stack named "' + Name + '" in the stack list.'#13#10'Use the edit function to modify the existing stack, or enter a new name.', mtError, [mbAbort,mbRetry], 0 ) = mrAbort;
      end
      else begin
        DeckFromString( NewDeckOrder, @Deck );
        FormEditor.SetDeck( Deck );
        if( FormEditor.ShowModal = mrOk ) then begin
          FormEditor.GetDeck( Deck );
          N := ComboStack.Items.Add( Name );
          FPresetDecks.Insert( N, deckToString( Deck ) );
          ComboStack.ItemIndex := N;
          ComboStackChange(Sender);
        end;
      end;
    end;
  until( Done );
  FormGetStackName.Free;
end;

procedure TFormMain.CmdEditStackClick(Sender: TObject);
begin
  FormEditor.SetDeck( FDeck );
  if( FormEditor.ShowModal = mrOk ) then begin
    FormEditor.GetDeck( FDeck );
    FPresetDecks[ ComboStack.ItemIndex ] := deckToString( FDeck );
    ComboStackChange(Sender);
  end;
end;

procedure TFormMain.CmdDeleteStackClick(Sender: TObject);
begin
  if( MessageDlg( 'Are you sure you want to delete the stack named "' + ComboStack.Items[ComboStack.ItemIndex] + '"?', mtConfirmation, mbYesNoCancel, 0 ) = mrYes ) then begin
    FPresetDecks.Delete( ComboStack.ItemIndex );
    ComboStack.Items.Delete( ComboStack.ItemIndex );
    CheckComboStackNotEmpty;
    ComboStack.ItemIndex := 0;
    ComboStackChange(Sender);
  end;
end;

procedure TFormMain.CmdAboutClick(Sender: TObject);
begin
  FormAbout := TFormAbout.Create(Self);
  FormAbout.ShowModal;
  FormAbout.Free;
end;

procedure TFormMain.CheckShuffledClick(Sender: TObject);
begin
  PaintCurrentCard;
end;

procedure TFormMain.CmdShuffleClick(Sender: TObject);
begin
  Shuffle;
  PaintCurrentCard;
end;

procedure TFormMain.CmdProgramHelpClick(Sender: TObject);
begin
  OpenHelp( Self.Handle, 'index' );
end;

procedure TFormMain.FormKeyDown(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  if( Shift = [] )and( Key = VK_F1 ) then begin
    OpenHelp( Self.Handle, 'main' );
  end;
end;

end.
